import numpy as np
import re
import tensorflow as tf


class CreateTensorFlowDataset:
    def __init__(self):
        self.dataset_map_options = {
            "num_parallel_calls": tf.data.experimental.AUTOTUNE,
            "deterministic": False,
        }

    def __call__(self, study_description):
        """From scratch, creates a tensorflow dataset with one tensorflow element per tile"""

        if not (
            "version" in study_description
            and study_description["version"] == "version-1"
        ):
            raise ValueError(
                'study_description["version"] must exist and be equal to "version-1".'
            )
        if not (
            "number_pixel_rows_for_tile" in study_description
            and isinstance(study_description["number_pixel_rows_for_tile"], int)
            and study_description["number_pixel_rows_for_tile"] > 0
        ):
            raise ValueError(
                'study_description["number_pixel_rows_for_tile"]'
                " must exist and be a positive integer"
            )
        if not (
            "number_pixel_columns_for_tile" in study_description
            and isinstance(study_description["number_pixel_columns_for_tile"], int)
            and study_description["number_pixel_columns_for_tile"] > 0
        ):
            raise ValueError(
                'study_description["number_pixel_columns_for_tile"]'
                " must exist and be a positive integer"
            )
        # Check that other necessary keys are also present!!!

        # Partition the set of tiles into chunks.
        self._designate_chunks_for_tiles(study_description)
        # cProfile.runctx("self._designate_chunks_for_tiles(study_description)", globals=globals(), locals=locals(), sort="cumulative")
        # print("_designate_chunks_for_tiles done")

        self.number_pixel_rows_for_tile = tf.convert_to_tensor(
            study_description["number_pixel_rows_for_tile"]
        )
        self.number_pixel_columns_for_tile = tf.convert_to_tensor(
            study_description["number_pixel_columns_for_tile"]
        )

        # Start converting our description into tensors.
        study_as_tensors = {
            study_key: [tf.convert_to_tensor(study_description[study_key])]
            for study_key in study_description.keys()
            if study_key != "slides"
        }
        # print("study_as_tensors done")

        number_of_chunks = 0
        for slide_description in study_description["slides"].values():
            slide_as_tensors = {
                **study_as_tensors,
                **{
                    slide_key: [tf.convert_to_tensor(slide_description[slide_key])]
                    for slide_key in slide_description.keys()
                    if slide_key != "chunks"
                },
            }

            for chunk_description in slide_description["chunks"].values():
                chunk_as_tensors = {
                    **slide_as_tensors,
                    **{
                        chunk_key: [tf.convert_to_tensor(chunk_description[chunk_key])]
                        for chunk_key in chunk_description.keys()
                        if chunk_key != "tiles"
                    },
                    "tiles_top": [
                        tf.convert_to_tensor(
                            [
                                tile["tile_top"]
                                for tile in chunk_description["tiles"].values()
                            ]
                        )
                    ],
                    "tiles_left": [
                        tf.convert_to_tensor(
                            [
                                tile["tile_left"]
                                for tile in chunk_description["tiles"].values()
                            ]
                        )
                    ],
                }

                # Make a tensorflow Dataset from this chunk.
                chunk_dataset = tf.data.Dataset.from_tensor_slices(chunk_as_tensors)
                if number_of_chunks == 0:
                    study_dataset = chunk_dataset
                else:
                    study_dataset = study_dataset.concatenate(chunk_dataset)
                number_of_chunks += 1

        # We have accumulated the chunk datasets into a study_dataset where each element
        # is a chunk.  Read in the chunk pixel data and split it into tiles.
        study_dataset = study_dataset.map(
            self._read_and_split_chunk_pixels, **self.dataset_map_options
        )
        # print("_read_and_split_chunk_pixels done")
        # Change study_dataset so that each element is a tile.
        study_dataset = study_dataset.unbatch()
        # print("unbatch done")

        # Make the tile pixels easier to find in each study_dataset element.  Also, tack
        # on additional elements to the tuple so that the form is (inputs, targets,
        # sample_weights).
        study_dataset = study_dataset.map(
            lambda elem: ((elem.pop("tile_pixels"), elem), None, None), **self.dataset_map_options
        )
        # print("pop done")

        return study_dataset

    def _designate_chunks_for_tiles(self, study_description):
        number_pixel_rows_for_tile = study_description["number_pixel_rows_for_tile"]
        number_pixel_columns_for_tile = study_description[
            "number_pixel_columns_for_tile"
        ]

        for slide in study_description["slides"].values():
            if not (
                "number_pixel_rows_for_chunk" in slide
                and isinstance(slide["number_pixel_rows_for_chunk"], int)
                and slide["number_pixel_rows_for_chunk"] > 0
            ):
                raise ValueError(
                    'slide["number_pixel_rows_for_chunk"]'
                    " must exist and be a positive integer"
                )
            if not (
                "number_pixel_columns_for_chunk" in slide
                and isinstance(slide["number_pixel_columns_for_chunk"], int)
                and slide["number_pixel_columns_for_chunk"] > 0
            ):
                raise ValueError(
                    'slide["number_pixel_columns_for_chunk"]'
                    " must exist and be a positive integer"
                )
            number_pixel_rows_for_chunk = slide["number_pixel_rows_for_chunk"]
            number_pixel_columns_for_chunk = slide["number_pixel_columns_for_chunk"]

            tiles_as_sorted_list = list(slide["tiles"].items())
            del slide["tiles"]
            tiles_as_sorted_list.sort(
                key=lambda x: x[1]["tile_left"]
            )  # second priority key
            tiles_as_sorted_list.sort(
                key=lambda x: x[1]["tile_top"]
            )  # first priority key
            chunks = slide["chunks"] = {}
            number_of_chunks = 0
            while len(tiles_as_sorted_list) > 0:
                tile = tiles_as_sorted_list[0]
                chunk = chunks[f"chunk_{number_of_chunks}"] = {
                    "chunk_top": tile[1]["tile_top"],
                    "chunk_left": tile[1]["tile_left"],
                    "chunk_bottom": tile[1]["tile_top"] + number_pixel_rows_for_chunk,
                    "chunk_right": tile[1]["tile_left"]
                    + number_pixel_columns_for_chunk,
                }
                number_of_chunks += 1

                if True:
                    # This implementations has a run time that is quadratic in the
                    # number of tiles that a slide has.  It is too slow; we should make
                    # it faster.
                    tiles = chunk["tiles"] = {}
                    subsequent_chunks = []
                    for tile in tiles_as_sorted_list:
                        if (
                            tile[1]["tile_top"] + number_pixel_rows_for_tile
                            <= chunk["chunk_bottom"]
                            and tile[1]["tile_left"] + number_pixel_columns_for_tile
                            <= chunk["chunk_right"]
                            and tile[1]["tile_left"] >= chunk["chunk_left"]
                            and tile[1]["tile_top"] >= chunk["chunk_top"]
                        ):
                            tiles[tile[0]] = tile[1]
                        else:
                            subsequent_chunks.append(tile)

                else:
                    # This implementations has a run time that is quadratic in the
                    # number of tiles that a slide has.  It is even slower than the
                    # above.
                    tiles = chunk["tiles"] = {
                        tile[0]: tile[1]
                        for tile in tiles_as_sorted_list
                        if tile[1]["tile_top"] + number_pixel_rows_for_tile
                        <= chunk["chunk_bottom"]
                        and tile[1]["tile_left"] + number_pixel_columns_for_tile
                        <= chunk["chunk_right"]
                        and tile[1]["tile_left"] >= chunk["chunk_left"]
                        and tile[1]["tile_top"] >= chunk["chunk_top"]
                    }
                    subsequent_chunks = [
                        tile
                        for tile in tiles_as_sorted_list
                        if not (
                            tile[1]["tile_top"] + number_pixel_rows_for_tile
                            <= chunk["chunk_bottom"]
                            and tile[1]["tile_left"] + number_pixel_columns_for_tile
                            <= chunk["chunk_right"]
                            and tile[1]["tile_left"] >= chunk["chunk_left"]
                            and tile[1]["tile_top"] >= chunk["chunk_top"]
                        )
                    ]

                # Update the list of tiles that are not yet in chunks
                tiles_as_sorted_list = subsequent_chunks

                # Make the chunk as small as possible given the tiles that it must
                # support.  Note that this also ensures that the pixels that are read do
                # not run over the bottom or right border of the slide (assuming that
                # the tiles do not go over those borders).
                chunk["chunk_top"] = min([tile["tile_top"] for tile in tiles.values()])
                chunk["chunk_left"] = min(
                    [tile["tile_left"] for tile in tiles.values()]
                )
                chunk["chunk_bottom"] = (
                    max([tile["tile_top"] for tile in tiles.values()])
                    + number_pixel_rows_for_tile
                )
                chunk["chunk_right"] = (
                    max([tile["tile_left"] for tile in tiles.values()])
                    + number_pixel_columns_for_tile
                )

    @tf.function
    def _read_and_split_chunk_pixels(self, elem):
        # Get chunk's pixel data from disk and load it into chunk_pixels_as_tensor
        chunk_pixels_as_tensor = tf.py_function(
            func=self._py_read_chunk_pixels,
            inp=[
                elem["chunk_top"],
                elem["chunk_left"],
                elem["chunk_bottom"],
                elem["chunk_right"],
                elem["filename"],
                elem["level"],
            ],
            Tout=tf.uint8,
        )
        number_of_tiles = tf.size(elem["tiles_top"])
        tiles = tf.TensorArray(dtype=tf.uint8, size=number_of_tiles)

        def condition(i, _):
            return tf.less(i, number_of_tiles)

        def body(i, tiles):
            return (
                i + 1,
                tiles.write(
                    i,
                    tf.image.crop_to_bounding_box(
                        chunk_pixels_as_tensor,
                        tf.gather(elem["tiles_top"], i) - elem["chunk_top"],
                        tf.gather(elem["tiles_left"], i) - elem["chunk_left"],
                        elem["number_pixel_rows_for_tile"],
                        elem["number_pixel_columns_for_tile"],
                    ),
                ),
            )

        _, tiles = tf.while_loop(condition, body, [0, tiles])
        tiles = tiles.stack()

        response = {}
        for key in elem.keys():
            if key not in ("tiles_top", "tiles_left"):
                response[key] = tf.repeat(elem[key], number_of_tiles)

        response = {
            **response,
            "tile_top": elem["tiles_top"],
            "tile_left": elem["tiles_left"],
            "tile_pixels": tiles,
        }
        return response

    def _py_read_chunk_pixels(
        self, chunk_top, chunk_left, chunk_bottom, chunk_right, filename, level=-1
    ):
        """Read from disk all the pixel data for a specific chunk of the whole slide."""

        filename = filename.numpy().decode("utf-8")
        chunk_top = chunk_top.numpy()
        chunk_left = chunk_left.numpy()
        chunk_bottom = chunk_bottom.numpy()
        chunk_right = chunk_right.numpy()
        level = level.numpy()

        if re.compile(r"\.svs$").search(filename):
            import large_image

            ts = large_image.open(filename)
            chunk = ts.getRegion(
                scale=dict(level=level),
                format=large_image.constants.TILE_FORMAT_NUMPY,
                region=dict(
                    left=chunk_left,
                    top=chunk_top,
                    width=chunk_right - chunk_left,
                    height=chunk_bottom - chunk_top,
                    units="mag_pixels",
                ),
            )[0]
        elif re.compile(r"\.svs$").search(filename):
            import openslide as os

            os_obj = os.OpenSlide(filename)
            chunk = np.array(
                os_obj.read_region(
                    (chunk_left, chunk_top),
                    level,
                    (chunk_right - chunk_left, chunk_bottom - chunk_top),
                )
            )
        else:
            from PIL import Image

            pil_obj = Image.open(filename)
            chunk = np.asarray(pil_obj)[
                chunk_left:chunk_right, chunk_top:chunk_bottom, :
            ]

        # Do we want to support other than RGB and/or other than uint8?!!!
        return tf.convert_to_tensor(chunk[..., :3], dtype=tf.uint8)



#!/usr/bin/env python3

# Copyright NumFOCUS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import histomics_stream as hs

def test_histomics_stream_can_be_found():
    """Purpose: Exercise the basic steps for creating a tensorflow Dataset"""


    # Create a study and insert study-wide information
    my_study0 = {"version": "version-1"}
    my_study0["number_pixel_rows_for_tile"] = 256
    my_study0["number_pixel_columns_for_tile"] = 256
    my_slides = my_study0["slides"] = {}

    # Add a slide to the study, including slide-wide information with it.
    my_slide0 = my_slides["Slide_0"] = {}
    my_slide0["filename"] = "/tf/notebooks/histomics_stream/example/TCGA-BH-A0BZ-01Z-00-DX1.45EB3E93-A871-49C6-9EAE-90D98AE01913.svs"
    my_slide0["slide_name"] = "TCGA-BH-A0BZ-01Z-00-DX1"
    my_slide0["slide_group"] = "TCGA-BH-A0BZ"
    my_slide0["number_pixel_rows_for_chunk"] = 2048
    my_slide0["number_pixel_columns_for_chunk"] = 2048

    if False:
        # For each slide, find the appropriate resolution given the desired_magnification and
        # magnification_tolerance.  In this example, we use the same parameters for each slide,
        # but this is not required generally.
        find_resolution_for_slide = hs.configure.FindResolutionForSlide(
            my_study0, desired_magnification=20, magnification_tolerance=0.02
        )
        for slide in my_study0["slides"].values():
            find_resolution_for_slide(slide)
    else:
        # Because we don't actually have the image available, make up some numbers.
        my_slide0["level"] = 0
        my_slide0["factor"] = 0.5
        my_slide0["number_pixel_rows_for_slide"] = 85047
        my_slide0["number_pixel_columns_for_slide"] = 112334

    # We are going to demonstrate several approaches to choosing tiles.  Each approach will
    # start with its own copy of the my_study0 that we have built so far.

    # Demonstrate TilesByGridAndMask without a mask
    my_study_tiles_by_grid = copy.deepcopy(my_study0)
    tiles_by_grid = hs.configure.TilesByGridAndMask(
        my_study_tiles_by_grid,
        number_pixel_overlap_rows_for_tile=32,
        number_pixel_overlap_columns_for_tile=32,
        randomly_select=100,
    )
    # We could apply this to a subset of the slides, but we will apply it to all slides in
    # this example.
    for slide in my_study_tiles_by_grid["slides"].values():
        tiles_by_grid(slide)

    if False:
        # Skip this test for now because we don't have the mask file available.
        # Demonstrate TilesByGridAndMask with a mask
        my_study_tiles_by_grid_and_mask = copy.deepcopy(my_study0)
        tiles_by_grid_and_mask = hs.configure.TilesByGridAndMask(
            my_study_tiles_by_grid_and_mask,
            number_pixel_overlap_rows_for_tile=0,
            number_pixel_overlap_columns_for_tile=0,
            mask_filename="/tf/notebooks/histomics_stream/example/TCGA-BH-A0BZ-01Z-00-DX1.45EB3E93-A871-49C6-9EAE-90D98AE01913-mask.png",
            randomly_select=100,
        )
        # We could apply this to a subset of the slides, but we will apply it to all slides in
        # this example.
        for slide in my_study_tiles_by_grid_and_mask["slides"].values():
            tiles_by_grid_and_mask(slide)

    # Demonstrate TilesByList
    my_study_tiles_by_list = copy.deepcopy(my_study0)
    tiles_by_list = hs.configure.TilesByList(
        my_study_tiles_by_list,
        randomly_select=5,
        tiles_dictionary=my_study_tiles_by_grid["slides"]["Slide_0"]["tiles"],
    )
    # We could apply this to a subset of the slides, but we will apply it to all slides in
    # this example.
    for slide in my_study_tiles_by_list["slides"].values():
        tiles_by_list(slide)

    # Demonstrate TilesRandomly
    my_study_tiles_randomly = copy.deepcopy(my_study0)
    tiles_randomly = hs.configure.TilesRandomly(
        my_study_tiles_randomly, randomly_select=3
    )
    # We could apply this to a subset of the slides, but we will apply it to all slides in
    # this example.
    for slide in my_study_tiles_randomly["slides"].values():
        tiles_randomly(slide)



"""Whole-slide image streamer for TensorFlow."""

__version__ = "2.1.1"

"""

This module supports efficient whole-slide reading and processing in a tensorflow graph.

"""
from . import configure, tensorflow, codecs



"""Whole-slide image file reader for TensorFlow.

The histomics_stream.codecs module supplies codecs that are useful for Zarr file storage with jpeg or
jpeg2k compression.

"""

from imagecodecs import jpeg_encode, jpeg_decode, jpeg2k_encode, jpeg2k_decode
from numcodecs.abc import Codec
from numcodecs.compat import ensure_ndarray, ensure_contiguous_ndarray, ndarray_copy
from numcodecs.registry import register_codec


class jpeg(Codec):
    """Codec providing jpeg compression via imagecodecs.

    Parameters
    ----------
    quality : int
        Compression level.

    Notes
    -----
    For the code that uses Zarr data storage for jpeg images, we need to supply codecs.  Note that we
    use this codec instead of that available from the zarr_jpeg package.  The latter collapses
    dimensions by default, can require us to transpose dimensions, and can miss optimizations based upon
    RGB data.

    """

    codec_id = "jpeg"

    def __init__(self, quality=100):
        self.quality = quality
        assert 0 < self.quality <= 100 and isinstance(self.quality, int)
        super().__init__()

    def encode(self, buf):
        """The method to encode a raw image into jpeg format.

        Parameters
        ----------
        buf : ndarray
            The raw image to be encoded into jpeg format

        Returns
        -------
        ndarray
            The image in jpeg format

        """

        bufa = ensure_ndarray(buf)
        assert 2 <= bufa.ndim <= 3
        return jpeg_encode(bufa, level=self.quality)

    def decode(self, buf, out=None):
        """The method to decode a jpeg image into a raw format.

        Parameters
        ----------
        buf : contiguous_ndarray
            The jpeg image to be decoded into raw format.
        out : contiguous_ndarray, optional
            Another location to write the raw image to.

        Returns
        -------
        ndarray
            The image in raw format

        """

        buf = ensure_contiguous_ndarray(buf)
        if out is not None:
            out = ensure_contiguous_ndarray(out)
        tiled = jpeg_decode(buf)
        return ndarray_copy(tiled, out)


register_codec(jpeg)


class jpeg2k(Codec):
    """Codec providing jpeg2k compression via imagecodecs.

    Parameters
    ----------
    quality : int
        Compression level.

    """

    codec_id = "jpeg2k"

    def __init__(self, quality=100):
        self.quality = quality
        assert 0 < self.quality <= 100 and isinstance(self.quality, int)
        super().__init__()

    def encode(self, buf):
        """The method to encode a raw image into jpeg2k format.

        Parameters
        ----------
        buf : ndarray
            The raw image to be encoded into jpeg2k format

        Returns
        -------
        ndarray
            The image in jpeg2k format

        """

        bufa = ensure_ndarray(buf)
        assert 2 <= bufa.ndim <= 3
        return jpeg2k_encode(bufa, level=self.quality)

    def decode(self, buf, out=None):
        """The method to decode a jpeg2k image into a raw format.

        Parameters
        ----------
        buf : contiguous_ndarray
            The jpeg2k image to be decoded into raw format.
        out : contiguous_ndarray, optional
            Another location to write the raw image to.

        Returns
        -------
        ndarray
            The image in raw format

        """

        buf = ensure_contiguous_ndarray(buf)
        if out is not None:
            out = ensure_contiguous_ndarray(out)
        tiled = jpeg2k_decode(buf)
        return ndarray_copy(tiled, out)


register_codec(jpeg2k)



import copy
import itk
import math
import numpy as np
import random
import re


class FindResolutionForSlide:
    """A class that computes read parameters for slides.

    An instance of class FindResolutionForSlide is a callable that will add level,
    factor, number_pixel_rows_for_slide, and number_pixel_columns_for_slide fields to a
    slide dictionary.

    Parameters for the constructor
    ------------------------------
    filename : string
        The path of the image file to be read.
    desired_magnification : float
        The desired magnification to be read from the file when
        multiple resolutions are available.  For example, a value of
        10 corresponses to about 1 micron per pixel and a value of 20
        corresponds to about 0.5 microns per pixel.
    magnification_tolerance : float
        For example a value of 0.02 allows selection of an image that
        has a magnification of up to 2% different from the desired
        magnification.  If no such image exists, the image with
        greater magnification will be selected.

    """

    def __init__(self, study, desired_magnification, magnification_tolerance):
        """Sanity check the supplied parameters and store them for later use."""
        # Check values.
        if not ("version" in study and study["version"] == "version-1"):
            raise ValueError('study["version"] must exist and be equal to "version-1".')
        if not (
            isinstance(desired_magnification, (int, float))
            and 0 < desired_magnification
        ):
            raise ValueError(
                f"desired_magnification ({desired_magnification})"
                " must be a positive number"
            )
        if not (
            isinstance(magnification_tolerance, (int, float))
            and 0 <= magnification_tolerance <= 1
        ):
            raise ValueError(
                f"magnification_tolerance ({magnification_tolerance})"
                " must be a value in [0, 1]"
            )

        # Save values.
        self.desired_magnification = desired_magnification
        self.magnification_tolerance = magnification_tolerance

    def __call__(self, slide):
        """Add level, factor, number_pixel_rows_for_slide, and number_pixel_columns_for_slide
        fields to a slide dictionary.

        """

        # Check values.
        if "filename" not in slide:
            raise ValueError('slide["filename"] must be already set.')
        filename = slide["filename"]

        # Do the work.
        if re.compile(r"\.svs$").search(filename):
            import large_image

            # read whole-slide image file and create large_image object
            ts = large_image.open(filename)

            # measure objective of level 0
            objective = np.float32(ts.getNativeMagnification()["magnification"])

            if False:
                # Use the level that large_image is willing to interpolate for us.
                preferred_levels = [
                    ts.getLevelForMagnification(
                        self.desired_magnification, rounding=False
                    )
                ]
            else:
                # Use one of the levels that is stored in the image file.
                preferred_levels = list(
                    set(ts.getPreferredLevel(level) for level in range(ts.levels))
                )
                preferred_levels.sort(reverse=True)

            estimated = np.array(
                [
                    ts.getMagnificationForLevel(level)["magnification"]
                    for level in preferred_levels
                ]
            )

            # Find best native level to use and its factor
            level, factor = self._get_level_and_factor(
                self.desired_magnification, estimated, self.magnification_tolerance
            )

            number_pixel_columns_for_slide = int(
                ts.sizeX * estimated[level] // objective
            )
            number_pixel_rows_for_slide = int(ts.sizeY * estimated[level] // objective)

            # Rather than as the index into preferred_levels, change
            # level to be the value that large_image uses
            level = preferred_levels[level]

        elif re.compile(r"\.svs$").search(filename):
            import openslide as os

            # read whole-slide image file and create openslide object
            os_obj = os.OpenSlide(filename)

            # measure objective of level 0
            objective = np.float32(os_obj.properties[os.PROPERTY_NAME_OBJECTIVE_POWER])

            # calculate magnifications of levels
            estimated = np.array(objective / os_obj.level_downsamples)

            # Find best native level to use and its factor
            level, factor = self._get_level_and_factor(
                self.desired_magnification, estimated, self.magnification_tolerance
            )

            # get slide number_pixel_columns_for_slide,
            # number_pixel_rows_for_slide at desired
            # magnification. (Note number_pixel_columns_for_slide
            # before number_pixel_rows_for_slide)
            (
                number_pixel_columns_for_slide,
                number_pixel_rows_for_slide,
            ) = os_obj.level_dimensions[level]

        elif re.compile(r"\.zarr$").search(filename):
            import zarr

            # read whole-slide image and create zarr objects
            store = zarr.DirectoryStore(filename)
            source_group = zarr.open(store, mode="r")

            # measure objective of level 0
            objective = np.float32(source_group.attrs[os.PROPERTY_NAME_OBJECTIVE_POWER])

            # calculate magnifications of levels
            estimated = np.array(objective / source_group.attrs["level_downsamples"])

            # Find best native level to use and its factor
            level, factor = self._get_level_and_factor(
                self.desired_magnification, estimated, self.magnification_tolerance
            )

            # get slide number_pixel_columns_for_slide,
            # number_pixel_rows_for_slide at desired
            # magnification. (Note number_pixel_rows_for_slide before
            # number_pixel_columns_for_slide)
            number_pixel_rows_for_slide, number_pixel_columns_for_slide = source_group[
                format(level)
            ].shape[0:2]

        else:
            from PIL import Image

            # We don't know magnifications so assume reasonable values
            # for level and factor.
            level = 0
            factor = 1.0
            pil_obj = Image.open(filename)
            number_pixel_columns_for_slide, number_pixel_rows_for_slide = pil_obj.size

        slide["level"] = level
        slide["factor"] = factor
        slide["number_pixel_rows_for_slide"] = number_pixel_rows_for_slide
        slide["number_pixel_columns_for_slide"] = number_pixel_columns_for_slide

    def _get_level_and_factor(
        self, desired_magnification, estimated, magnification_tolerance
    ):
        """A private subroutine that computes level and factor."""
        # calculate difference with magnification levels
        delta = desired_magnification - estimated

        # match to existing levels
        if (
            np.min(np.abs(np.divide(delta, desired_magnification)))
            < magnification_tolerance
        ):  # match
            level = np.squeeze(np.argmin(np.abs(delta)))
            factor = 1.0
        elif np.any(delta < 0):
            value = np.max(delta[delta < 0])
            level = np.squeeze(np.argwhere(delta == value)[0])
            factor = desired_magnification / estimated[level]
        else:  # desired magnification above base level - throw error
            raise ValueError("Cannot interpolate above scan magnification.")

        return level, factor


class TilesByGridAndMask:
    """Select tiles according to a regular grid.  Optionally, restrict
    the list by a mask that is read from a file.  Optionally, further
    select a random subset of them.

    An instance of class TilesByGridAndMask is a callable that will
    select the coordinates of tiles to be taken from a slide.  The
    selected tiles will be written to the slide dictionary.

    Parameters for the constructor
    ------------------------------
    study : dictionary
        The study dictionary from which to read parameters about the study.
    randomly_select: int
        The number of tiles to be randomly selected from the list that
        would otherwise be written to the slide dictionary.  A value
        of -1 is the default and means that all tiles should be
        written.
    number_pixel_overlap_rows_for_tile
        Specifies the desired amount of vertical overlap (measured in
        rows of pixels) between adjacent tiles.  This defaults to 0,
        which means that there is no overlap between adjacent tiles;
        they are abutting.
    number_pixel_overlap_columns_for_tile
        Specifies the desired amount of horizontal overlap (measured
        in columns of pixels) between adjacent tiles.  This defaults
        to 0, which means that there is no overlap between adjacent
        tiles; they are abutting.
    mask_filename: string
        The path of the image file to be read and used as a mask.  The
        aspect ratio of the mask (in terms of its pixel dimensions) is
        expected to be about the same as the aspect ratio of the main
        image ( in terms of its grid of tiles).  A non-zero value in
        the mask indicates that the tile should be retained.  The
        default is "", which means that there is no masking.

    """

    def __init__(
        self,
        study,
        randomly_select=-1,  # Defaults to select all
        number_pixel_overlap_rows_for_tile=0,  # Defaults to no overlap between adjacent tiles
        number_pixel_overlap_columns_for_tile=0,
        mask_filename="",  # Defaults to no masking
    ):
        """Sanity check the supplied parameters and store them for later use."""
        # Check values.
        if not ("version" in study and study["version"] == "version-1"):
            raise ValueError('study["version"] must exist and be equal to "version-1".')
        if not (
            "number_pixel_rows_for_tile" in study
            and isinstance(study["number_pixel_rows_for_tile"], int)
            and study["number_pixel_rows_for_tile"] > 0
        ):
            raise ValueError(
                'study["number_pixel_rows_for_tile"]'
                " must exist and be a positive integer"
            )
        if not (
            "number_pixel_columns_for_tile" in study
            and isinstance(study["number_pixel_columns_for_tile"], int)
            and study["number_pixel_columns_for_tile"] > 0
        ):
            raise ValueError(
                'study["number_pixel_columns_for_tile"]'
                " must exist and be a positive integer"
            )
        if not (isinstance(randomly_select, int) and -1 <= randomly_select):
            raise ValueError(
                f"randomly_select ({randomly_select})"
                " must be a non-negative integer or -1."
            )
        if not (
            isinstance(number_pixel_overlap_rows_for_tile, int)
            and number_pixel_overlap_rows_for_tile < study["number_pixel_rows_for_tile"]
        ):
            raise ValueError(
                f"number_pixel_overlap_rows_for_tile ({number_pixel_overlap_rows_for_tile})"
                " must be less than"
                f' number_pixel_rows_for_tile ({study["number_pixel_rows_for_tile"]}).'
            )
        if not (
            isinstance(number_pixel_overlap_columns_for_tile, int)
            and number_pixel_overlap_columns_for_tile
            < study["number_pixel_columns_for_tile"]
        ):
            raise ValueError(
                f"number_pixel_overlap_columns_for_tile ({number_pixel_overlap_columns_for_tile})"
                " must be less than"
                f' number_pixel_columns_for_tile ({study["number_pixel_columns_for_tile"]}).'
            )
        if mask_filename != "":
            mask_itk = itk.imread(mask_filename)  # May throw exception
            if mask_itk.GetImageDimension() != 2:
                raise ValueError(
                    f"The mask ({mask_filename})" " should be a 2-dimensional image."
                )

        # Save values.  To keep garbage collection efficient don't save all of `study`.
        self.number_pixel_rows_for_tile = study["number_pixel_rows_for_tile"]
        self.number_pixel_columns_for_tile = study["number_pixel_columns_for_tile"]
        self.randomly_select = randomly_select
        self.number_pixel_overlap_rows_for_tile = number_pixel_overlap_rows_for_tile
        self.number_pixel_overlap_columns_for_tile = (
            number_pixel_overlap_columns_for_tile
        )
        self.mask_filename = mask_filename
        if mask_filename != "":
            self.mask_itk = mask_itk

    def __call__(self, slide):
        """Select tiles according to a regular grid.  Optionally, restrict the list by a mask.
        Optionally, select a random subset of them.
        """
        # Check values.
        if "number_pixel_rows_for_slide" not in slide:
            raise ValueError(
                'slide["number_pixel_rows_for_slide"] must be already set.'
            )
        if "number_pixel_columns_for_slide" not in slide:
            raise ValueError(
                'slide["number_pixel_columns_for_slide"] must be already set.'
            )

        # Do the work.
        row_stride = (
            self.number_pixel_rows_for_tile - self.number_pixel_overlap_rows_for_tile
        )
        number_tile_rows_for_slide = slide["number_tile_rows_for_slide"] = math.floor(
            (
                slide["number_pixel_rows_for_slide"]
                - self.number_pixel_overlap_rows_for_tile
            )
            / row_stride
        )
        column_stride = (
            self.number_pixel_columns_for_tile
            - self.number_pixel_overlap_columns_for_tile
        )
        number_tile_columns_for_slide = slide[
            "number_tile_columns_for_slide"
        ] = math.floor(
            (
                slide["number_pixel_columns_for_slide"]
                - self.number_pixel_overlap_columns_for_tile
            )
            / column_stride
        )
        has_mask = hasattr(self, "mask_itk")
        if has_mask:
            # We will change the resolution of the mask (if necessary), which will
            # change the number of pixels, but will not change the overall physical size
            # represented by the image nor the position of the upper left corner of its
            # upper left pixel.
            input_size = itk.size(self.mask_itk)
            output_size = [number_tile_columns_for_slide, number_tile_rows_for_slide]
            if input_size != output_size:
                # print(f"Resampling from input_size = {input_size} to output_size = {output_size}")
                # Check that the input and output aspect ratios are pretty close
                if (
                    abs(
                        math.log(
                            (output_size[0] / input_size[0])
                            / (output_size[1] / input_size[1])
                        )
                    )
                    > 0.20
                ):
                    raise ValueError(
                        "The mask aspect ratio does not match that for the number of tiles."
                    )
                input_spacing = itk.spacing(self.mask_itk)
                input_origin = itk.origin(self.mask_itk)
                image_dimension = self.mask_itk.GetImageDimension()
                output_spacing = [
                    input_spacing[d] * input_size[d] / output_size[d]
                    for d in range(image_dimension)
                ]
                output_origin = [
                    input_origin[d] + 0.5 * (output_spacing[d] - input_spacing[d])
                    for d in range(image_dimension)
                ]
                interpolator = itk.NearestNeighborInterpolateImageFunction.New(
                    self.mask_itk
                )
                resampled_mask_itk = itk.resample_image_filter(
                    self.mask_itk,
                    interpolator=interpolator,
                    size=output_size,
                    output_spacing=output_spacing,
                    output_origin=output_origin,
                )
            else:
                resampled_mask_itk = self.mask_itk

        tiles = slide["tiles"] = {}
        number_of_tiles = 0
        for row in range(number_tile_rows_for_slide):
            for column in range(number_tile_columns_for_slide):
                if not (has_mask and resampled_mask_itk[row, column] == 0):
                    tiles[f"tile_{number_of_tiles}"] = {
                        "tile_top": row * row_stride,
                        "tile_left": column * column_stride,
                    }
                number_of_tiles += 1  # Increment even if tile is skipped.
        # Choose a subset of the tiles randomly
        all_tile_names = tiles.keys()
        if 0 <= self.randomly_select < len(all_tile_names):
            keys_to_remove = random.sample(
                all_tile_names, len(all_tile_names) - self.randomly_select
            )
            for key in keys_to_remove:
                del tiles[key]


class TilesByList:
    """Select the tiles supplied by the user.  Optionally, select a
    random subset of them.

    An instance of class TilesByList is a callable that will select
    the coordinates of tiles to be taken from a slide.  The selected
    tiles will be written to the slide dictionary.

    Parameters for the constructor
    ------------------------------
    study : dictionary
        The study dictionary from which to read parameters about the study.
    randomly_select: int
        The number of tiles to be randomly selected from the list that
        would otherwise be written to the slide dictionary.  A value
        of -1 is the default and means that all tiles should be
        written.
    tiles_dictionary: dictionary
        For example, {'AB234': {'tile_top': top0, 'tile_left': left0},
        'CD43': {'tile_top': top1, 'tile_left': left1}, ...}.  Tiles
        from this list will copied into the slide dictionary if they
        are randomly selected.

    """

    def __init__(
        self,
        study,
        randomly_select=-1,  # Defaults to select all
        tiles_dictionary={},  # {'AB234': {'tile_top': top0, 'tile_left': left0}, 'CD43': {'tile_top': top1, 'tile_left': left1}, ...}
    ):
        """Sanity check the supplied parameters and store them for later use."""
        # Check values
        if not ("version" in study and study["version"] == "version-1"):
            raise ValueError('study["version"] must exist and be equal to "version-1".')
        if not (
            "number_pixel_rows_for_tile" in study
            and isinstance(study["number_pixel_rows_for_tile"], int)
            and study["number_pixel_rows_for_tile"] > 0
        ):
            raise ValueError(
                'study["number_pixel_rows_for_tile"]'
                " must exist and be a positive integer"
            )
        if not (
            "number_pixel_columns_for_tile" in study
            and isinstance(study["number_pixel_columns_for_tile"], int)
            and study["number_pixel_columns_for_tile"] > 0
        ):
            raise ValueError(
                'study["number_pixel_columns_for_tile"]'
                " must exist and be a positive integer"
            )
        if not (isinstance(randomly_select, int) and -1 <= randomly_select):
            raise ValueError(
                f"randomly_select ({randomly_select})"
                " must be a non-negative integer or -1."
            )
        if not (
            isinstance(tiles_dictionary, dict)
            and all(
                [
                    isinstance(tile_corner, dict)
                    for tile_corner in tiles_dictionary.values()
                ]
            )
            and all(
                [
                    key in tile_corner.keys()
                    for tile_corner in tiles_dictionary.values()
                    for key in ("tile_top", "tile_left")
                ]
            )
            and all(
                [
                    isinstance(tile_corner[key], int)
                    for tile_corner in tiles_dictionary.values()
                    for key in ("tile_top", "tile_left")
                ]
            )
            and all(
                [
                    tile_corner[key] >= 0
                    for tile_corner in tiles_dictionary.values()
                    for key in ("tile_top", "tile_left")
                ]
            )
        ):
            raise ValueError(
                "tiles_dictionary must be dictionary of tiles."
                '  Each tile is a dictionary, with keys "tile_top" and "tile_left"'
                " and with values that are non-negative integers."
            )

        # Save values.  To keep garbage collection efficient don't save all of `study`,
        # just the parts that we need.
        self.number_pixel_rows_for_tile = study["number_pixel_rows_for_tile"]
        self.number_pixel_columns_for_tile = study["number_pixel_columns_for_tile"]
        self.randomly_select = randomly_select
        self.tiles_dictionary = copy.deepcopy(
            tiles_dictionary
        )  # in case user changes it later

    def __call__(self, slide):
        """Select the tiles supplied by the user.  Optionally, select a random subset of them."""
        tiles = slide["tiles"] = copy.deepcopy(
            self.tiles_dictionary
        )  # in case __call__ is called again.
        all_tile_names = tiles.keys()
        if 0 <= self.randomly_select < len(all_tile_names):
            keys_to_remove = random.sample(
                all_tile_names, len(all_tile_names) - self.randomly_select
            )
            for key in keys_to_remove:
                del tiles[key]


class TilesRandomly:
    """Select a random subset of all possible tiles.

    An instance of class TilesRandomly is a callable that will select
    the coordinates of tiles to be taken from a slide.  The selected
    tiles will be written to the slide dictionary.

    Parameters for the constructor
    ------------------------------
    study : dictionary
        The study dictionary from which to read parameters about the study.
    randomly_select: int
        The number of tiles to be randomly selected from the slide.
        The value must be positive.  A value of 1 is the default.

    """

    def __init__(
        self,
        study,
        randomly_select=1,  # Defaults to select one
    ):
        """Sanity check the supplied parameters and store them for later use."""
        # Check values.
        if not ("version" in study and study["version"] == "version-1"):
            raise ValueError('study["version"] must exist and be equal to "version-1".')
        if not (
            "number_pixel_rows_for_tile" in study
            and isinstance(study["number_pixel_rows_for_tile"], int)
            and study["number_pixel_rows_for_tile"] > 0
        ):
            raise ValueError(
                'study["number_pixel_rows_for_tile"]'
                " must exist and be a positive integer"
            )
        if not (
            "number_pixel_columns_for_tile" in study
            and isinstance(study["number_pixel_columns_for_tile"], int)
            and study["number_pixel_columns_for_tile"] > 0
        ):
            raise ValueError(
                'study["number_pixel_columns_for_tile"]'
                " must exist and be a positive integer"
            )
        if not (isinstance(randomly_select, int) and 0 <= randomly_select):
            raise ValueError(
                f"randomly_select ({randomly_select})"
                " must be a non-negative integer."
            )

        # Save values.  To keep garbage collection efficient don't save all of `study`.
        self.number_pixel_rows_for_tile = study["number_pixel_rows_for_tile"]
        self.number_pixel_columns_for_tile = study["number_pixel_columns_for_tile"]
        self.randomly_select = randomly_select

    def __call__(self, slide):
        """Select a random subset of all possible tiles."""
        if "number_pixel_rows_for_slide" not in slide:
            raise ValueError(
                'slide["number_pixel_rows_for_slide"] must be already set.'
            )
        if "number_pixel_columns_for_slide" not in slide:
            raise ValueError(
                'slide["number_pixel_columns_for_slide"] must be already set.'
            )

        row_too_big = (
            slide["number_pixel_rows_for_slide"] - self.number_pixel_rows_for_tile + 1
        )
        column_too_big = (
            slide["number_pixel_columns_for_slide"]
            - self.number_pixel_columns_for_tile
            + 1
        )
        row_column_list = [
            (random.randrange(0, row_too_big), random.randrange(0, column_too_big))
            for _ in range(self.randomly_select)
        ]
        tiles = slide["tiles"] = {}
        number_of_tiles = 0
        for (row, column) in row_column_list:
            tiles[f"tile_{number_of_tiles}"] = {"tile_top": row, "tile_left": column}
            number_of_tiles += 1



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="histomics_stream",
    version="2.1.1",
    author="Lee Newberg",
    author_email="lee.newberg@kitware.com",
    description="A TensorFlow 2 package for reading whole slide images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DigitalSlideArchive/histomics_stream",
    packages=setuptools.find_packages(exclude=["tests"]),
    package_dir={
        "histomics_stream": "histomics_stream",
    },
    install_requires=[
        # tensorflow is not listed as a requirement because the user
        # will likely want to specify the version for compatibility
        # with the CUDA libraries, etc.
        "itk",
        "pillow",
        "imagecodecs",
        "openslide-python",
        "large-image[all]",
        "zarr",
        "napari_lazy_openslide",
        "tifffile",
    ],
    license="Apache Software License 2.0",
    keywords="histomics_stream",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Image Recognition",
    ],
    zip_safe=False,
    python_requires=">=3.6",
)
